/* $Id: cpl_fits.c,v 1.4 2008/07/21 08:48:40 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/07/21 08:48:40 $
 * $Revision: 1.4 $
 * $Name: cpl-4_2_0b3 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <fitsio.h>

#include "cpl_fits.h"
#include "cpl_error.h"
#include "cpl_memory.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_fits   FITS related basic routines
 *
 * This module provides functions to get basic information on FITS files
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_fits.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the number of extensions contained in a FITS file
  @param    filename    The file name
  @return   the number of extensions or -1 in error case

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if the input pointer is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input file is not FITS
 */
/*----------------------------------------------------------------------------*/
int cpl_fits_get_nb_extensions(const char * filename)
{
    fitsfile    *   fptr;
    int             fio_status = 0;
    int             next       = 0;

    cpl_ensure(filename, CPL_ERROR_NULL_INPUT, -1);

    fits_open_diskfile(&fptr, filename, READONLY, &fio_status);
    cpl_ensure(fio_status == 0, CPL_ERROR_ILLEGAL_INPUT, -1);

    if (fits_get_num_hdus(fptr, &next, &fio_status)) {
        fits_close_file(fptr, &fio_status) ;
	return -1;
    }

    fits_close_file(fptr, &fio_status) ;

    /* The nb of HDUs includes the primary one, which is not an extension */
    return next - 1;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the place of a given extensions in a FITS file
  @param    filename    The file name
  @param    extname     The extension name
  @return   the extension place or 0 if it is not there or -1 in error case

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if one of the input pointers is NULL
 */
/*----------------------------------------------------------------------------*/
int cpl_fits_get_extension_nb(
        const char  *   filename,
        const char  *   extname)
{
    fitsfile    *   fptr;
    int             fio_status  = 0;
    int             ext_num;
    char            keyword[80] = "EXTNAME";

    cpl_ensure(filename, CPL_ERROR_NULL_INPUT, -1);
    cpl_ensure(extname,  CPL_ERROR_NULL_INPUT, -1);

    fits_open_diskfile(&fptr, filename, READONLY, &fio_status);
    cpl_ensure(fio_status == 0, CPL_ERROR_ILLEGAL_INPUT, -1);

    ext_num = 0;
    while (fio_status == 0) {
        char c_extname[80];

        fits_movabs_hdu(fptr, ext_num + 2, NULL, &fio_status);

        if (fio_status == END_OF_FILE) {
            fits_close_file(fptr, &fio_status);
            return -1;
        }

        ext_num++;

        fits_read_key_str(fptr, keyword, c_extname, NULL, &fio_status);

        if(!strcmp(extname, c_extname)) break;
    }

    fits_close_file(fptr, &fio_status);

    return ext_num;
}

/**@}*/
