/* $Id: cpl_image_filter.c,v 1.73 2008/07/21 08:48:40 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/07/21 08:48:40 $
 * $Revision: 1.73 $
 * $Name: cpl-4_2_0b3 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include "cpl_memory.h"
#include "cpl_image_filter.h"
#include "cpl_image_bpm.h"
#include "cpl_mask.h"
#include "cpl_tools.h"

#include "cpl_image_defs.h"

/*-----------------------------------------------------------------------------
                                   Defines
 -----------------------------------------------------------------------------*/

#define CPL_IMAGE_FILTER_LINEAR_NOBPM           1
#define CPL_IMAGE_FILTER_LINEAR                 2
#define CPL_IMAGE_FILTER_MORPHO_NOBPM           3
#define CPL_IMAGE_FILTER_MORPHO                 4
#define CPL_IMAGE_FILTER_MEDIAN_NOBPM           5
#define CPL_IMAGE_FILTER_MEDIAN                 6
#define CPL_IMAGE_FILTER_STDEV_NOBPM            7
#define CPL_IMAGE_FILTER_STDEV                  8

#define CONCAT(a,b) a ## _ ## b
#define CONCAT2X(a,b) CONCAT(a,b)

/* Caveat: One of a and b is evaluated twice */
#define CPL_IMAGE_FILTER_MAX(a,b) ((a) > (b) ? (a) : (b))
#define CPL_IMAGE_FILTER_MIN(a,b) ((a) < (b) ? (a) : (b))

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_image_filter Image filtering functions
 *
 * This module provides functions to filter images.
 *
 * The filtering functions (linear, median, morphological filtering) take
 * the bad pixels map (bpm) into account to avoid them to corrupt their 
 * neigbours.
 * 
 * @par Synopsis:
 * @code
 *   #include "cpl_image_filter.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image * cpl_image_filter_linear_nobpm(const cpl_image *,
        const cpl_matrix *) ;
static cpl_image * cpl_image_filter_morpho_nobpm(const cpl_image *,
        const cpl_matrix *) ;
static cpl_image * cpl_image_filter_median_nobpm(const cpl_image *,
        const cpl_matrix *) ;
static cpl_image * cpl_image_filter_stdev_nobpm(const cpl_image *,
        const cpl_matrix *) ;
static cpl_error_code cpl_mask_flag_borders(cpl_mask *, int, int) ;

static double cpl_matrix_norm_abs(const cpl_matrix*);

static cpl_image * cpl_image_filter_median_uniform_double(const cpl_image *,
                                                          int, int);
static cpl_image * cpl_image_filter_median_uniform_float(const cpl_image *,
                                                         int, int);

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

#define CPL_OPERATION CPL_IMAGE_FILTER_LINEAR
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute a linear filtering
  @param    in      The image to filter
  @param    ker     The kernel 
  @return   The filtered image or NULL in error case
  
  The function takes the bad pixels in account to avoid bad pixels corrupting 
  their neighbours.
  
  The input kernel should have an odd number of rows and columns.
  The maximum size of the kernel in both x and y is 31.

  The kernel values are the weight applied for the linear filtering.

  @verbatim
  Kernel 1 2 3    Image        ...
         4 5 6         ... 1.0 2.0 3.0 ...
         7 8 9         ... 4.0 5.0 6.0 ...
                       ... 7.0 8.0 9.0 ...
                       3.14    ...
  @endverbatim

  The pixel whose value is 5.0 is replaced by:
  \f$\frac{(1*1.0+2*2.0+3*3.0+4*4.0+5*5.0+6*6.0+7*7.0+8*8.0+9*9.0)}{1+2+3+4+5+6+7+8+9}\f$
 
  Beware that the element are not internally stored in the same order in
  a matrix and in an image:
  The Image first element is 3.14
  The Matrix first element is 1

  The output image has to be deallocated with cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_filter_linear(
        const cpl_image     *   in,
        const cpl_matrix    *   ker)
{
    const double    *   ker_arr ;
    int                 nc, nr ;
    cpl_image       *   filtered ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpm_dil ;
    cpl_binary      *   pbpm ;
    cpl_binary      *   pbpm_dil ;
    double              sum_pix ;
    double              ker_norm ;
    int                 hsx, hsy ;
    int                 curr_pos, im_pos, filt_pos;
    int                 x, y ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in,  CPL_ERROR_NULL_INPUT, NULL) ;
    cpl_ensure(ker, CPL_ERROR_NULL_INPUT, NULL) ;
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr & 1,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc <= 31, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr <= 31, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker)>0,CPL_ERROR_ILLEGAL_INPUT, NULL);

    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;
    
    /* First apply the normal filtering - and flag the border as bad */
    filtered = cpl_image_filter_linear_nobpm(in, ker) ;
    cpl_ensure(filtered!=NULL, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    
    /* Create the bad pixels map */
    if (in->bpm == NULL)    bpm = cpl_mask_new(in->nx, in->ny) ;
    else                    bpm = cpl_mask_duplicate(in->bpm) ;
    pbpm = cpl_mask_get_data(bpm) ;
    
    /* Create a map to identify bad pixels regions */
    bpm_dil = filtered->bpm ;
    pbpm_dil = cpl_mask_get_data(bpm_dil) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_mask_delete(bpm) ;
            cpl_image_delete(filtered) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    cpl_mask_delete(bpm) ;

    /* FIXME: Rough estimate, assuming a full kernel evaluation on all pixels */
    cpl_tools_add_flops( (1+2*nr*nc) * ( in->nx * in->ny ) );

    /* FIXME: Find a better way to determine if the bpm is empty */
    if (filtered->bpm != NULL && cpl_mask_is_empty(filtered->bpm)) {
        cpl_mask_delete(filtered->bpm);
        filtered->bpm = NULL;
    }

    return filtered;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_MORPHO
/*----------------------------------------------------------------------------*/
/**
  @brief    Filter an image in spatial domain with a morpho kernel.
  @param    in          Image to filter.
  @param    ker         Filter definition.
  @return   1 newly allocated image or NULL in error case.
  @see      cpl_image_filter_linear

  The function take the bad pixels in account to avoid bad pixels corrupting 
  their neighbours.
  
  The input kernel should have an odd number of rows and columns.
  The input filter is defined by a square matrix whose size is smaller than 
  5x5. The first element is applied to the min pixel in the neighborhood, 
  the second to the second-to-min, etc. and the last coefficient is applied 
  to the max pixel.
  The returned image is a newly allocated object, it must be deallocated using
  cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_filter_morpho(
        const cpl_image  * in,
        const cpl_matrix * ker)
{
    cpl_image       *   filtered ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpm_dil ;
    cpl_binary      *   pbpm ;
    cpl_binary      *   pbpm_dil ;
    double              sum_pix ;
    double              ker_norm = 0;
    int                 curr_pos, im_pos, filt_pos ;
    int                 hsx, hsy ;
    int                 nr, nc ;
    const double    *   ker_arr;
    double          *   pix_arr ;
    int                 nb_good ;
    int                 x, y ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in && ker, CPL_ERROR_NULL_INPUT, NULL) ;
 
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1,   CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc <= 5,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc == nr, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker)>0,CPL_ERROR_ILLEGAL_INPUT, NULL);
    
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;
        
    /* First apply the normal filtering - and flag the border as bad */
    filtered = cpl_image_filter_morpho_nobpm(in, ker) ;
    cpl_ensure(filtered!=NULL, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    
    /* Create the bad pixels map */
    if (in->bpm == NULL)    bpm = cpl_mask_new(in->nx, in->ny) ;
    else                    bpm = cpl_mask_duplicate(in->bpm) ;
    pbpm = cpl_mask_get_data(bpm) ;
    
    /* Create a map to identify bad pixels regions */
    bpm_dil = filtered->bpm ;
    pbpm_dil = cpl_mask_get_data(bpm_dil) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_mask_delete(bpm) ;
            cpl_image_delete(filtered) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    cpl_mask_delete(bpm) ;

    /* FIXME: Find a better way to determine if the bpm is empty */
    if (filtered->bpm != NULL && cpl_mask_is_empty(filtered->bpm)) {
        cpl_mask_delete(filtered->bpm);
        filtered->bpm = NULL;
    }

    return filtered;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_MEDIAN
/*----------------------------------------------------------------------------*/
/**
  @brief    Apply a spatial median filter to an image
  @param    in      Image to filter.
  @param    ker     the kernel
  @return   1 newly allocated image or NULL in error case
  @see      cpl_image_filter_linear

  The function take the bad pixels in account to avoid bad pixels corrupting 
  their neighbours.
 
  The kernel values set to 1.0 define the pixels taken into account in
  the median computation

  @verbatim
  Kernel 1.0 0.0 1.0    Image        ...
         0.0 1.0 0.0         ... 1.0 2.0 3.0 ...
         1.0 0.0 1.0         ... 4.0 5.0 6.0 ...
                             ... 7.0 8.0 9.0 ...
                                     ...
  @endverbatim

  The pixel whose value is 5.0 is replaced by:
  median(1.0, 3.0, 5.0, 7.0, 9.0) = 5.0
 
  Apply a spatial median filter to an image, return a newly allocated image
  which must be deallocated using cpl_image_delete().
  The kernel number of elements must be lower than or equal to 81.
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_filter_median(
        const cpl_image  * in,
        const cpl_matrix * ker)
{
    cpl_image       *   filtered ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpm_dil ;
    cpl_binary      *   pbpm ;
    cpl_binary      *   pbpm_dil ;
    int                 curpos, im_pos ;
    int                 nb_good ;
    const double    *   ker_arr ;
    int                 hsx, hsy ;
    int                 nr, nc ;
    int                 x, y ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in,  CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(ker, CPL_ERROR_NULL_INPUT, NULL);
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;

    /* Test the kernel validity */
    cpl_ensure(nr*nc<=81, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker)>0,CPL_ERROR_ILLEGAL_INPUT, NULL);

    cpl_ensure(in->type == CPL_TYPE_DOUBLE || in->type == CPL_TYPE_FLOAT,
               CPL_ERROR_TYPE_MISMATCH, NULL);
        
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;

    if (in->bpm == NULL && (nc & 1) && (nr & 1) &&
        cpl_matrix_get_min(ker) == 1.0 &&
        cpl_matrix_get_max(ker) == 1.0) {

        filtered = in->type == CPL_TYPE_DOUBLE
            ? cpl_image_filter_median_uniform_double(in, hsx, hsy)
            : cpl_image_filter_median_uniform_float(in, hsx, hsy);

        cpl_ensure(filtered != NULL, cpl_error_get_code(), NULL);

        return filtered;
    }
    
    /* First apply the normal filtering - and flag the border as bad */
    filtered = cpl_image_filter_median_nobpm(in, ker) ;
    cpl_ensure(filtered!=NULL, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    
    /* Create the bad pixels map */
    if (in->bpm == NULL)    bpm = cpl_mask_new(in->nx, in->ny) ;
    else                    bpm = cpl_mask_duplicate(in->bpm) ;
    pbpm = cpl_mask_get_data(bpm) ;
    
    /* Create a map to identify bad pixels regions */
    bpm_dil = filtered->bpm ;
    pbpm_dil = cpl_mask_get_data(bpm_dil) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_mask_delete(bpm) ;
            cpl_image_delete(filtered) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
            break ;
    }
    cpl_mask_delete(bpm) ;

    /* FIXME: Find a better way to determine if the bpm is empty */
    if (filtered->bpm != NULL && cpl_mask_is_empty(filtered->bpm)) {
        cpl_mask_delete(filtered->bpm);
        filtered->bpm = NULL;
    }

    return filtered;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_STDEV
/*----------------------------------------------------------------------------*/
/**
  @brief    Standard deviation filter 
  @param    in      input image
  @param    ker        the kernel
  @return   a newly allocated filtered image or NULL in error case
  @see      cpl_image_filter_linear
  
  The function take the bad pixels in account to avoid bad pixels corrupting 
  their neighbours.
 
  The kernel values set to 1.0 define the pixels taken into account in
  the computation

  @verbatim
  Kernel 1.0 0.0 1.0    Image        ...
         0.0 1.0 0.0         ... 1.0 2.0 3.0 ...
         1.0 0.0 1.0         ... 4.0 5.0 6.0 ...
                             ... 7.0 8.0 9.0 ...
                                     ...
  @endverbatim

  The pixel whose value is 5.0 is replaced by:
  std_dev(1.0, 3.0, 5.0, 7.0, 9.0)
 
  The input kernel should have an odd number of rows and columns.
  Apply a spatial standard deviation filter to an image, return a newly 
  allocated image which must be deallocated using cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_filter_stdev(
        const cpl_image  * in,
        const cpl_matrix * ker)
{
    cpl_image       *   filtered ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpm_dil ;
    cpl_binary      *   pbpm ;
    cpl_binary      *   pbpm_dil ;
    int                 curpos, im_pos ;
    const double    *   ker_arr ;
    int                 hsx, hsy ;
    int                 nr, nc ;
    register double     pix ;
    double              sum ;
    double              sq_sum ;
    double              inv_surf ;
    int                 x, y ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in,  CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(ker, CPL_ERROR_NULL_INPUT, NULL);
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr & 1, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker)>0,CPL_ERROR_ILLEGAL_INPUT, NULL);
        
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;

    /* First apply the normal stdev filtering - and flag the border as bad */
    filtered = cpl_image_filter_stdev_nobpm(in, ker) ;
    cpl_ensure(filtered!=NULL, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    
    /* Create the bad pixels map */
    if (in->bpm == NULL)    bpm = cpl_mask_new(in->nx, in->ny) ;
    else                    bpm = cpl_mask_duplicate(in->bpm) ;
    pbpm = cpl_mask_get_data(bpm) ;
    
    /* Create a map to identify bad pixels regions */
    bpm_dil = filtered->bpm ;
    pbpm_dil = cpl_mask_get_data(bpm_dil) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_mask_delete(bpm) ;
            cpl_image_delete(filtered) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL);
            break ;
    }
    cpl_mask_delete(bpm) ;

    /* FIXME: Find a better way to determine if the bpm is empty */
    if (filtered->bpm != NULL && cpl_mask_is_empty(filtered->bpm)) {
        cpl_mask_delete(filtered->bpm);
        filtered->bpm = NULL;
    }

    return filtered;
}
#undef CPL_OPERATION
/**@}*/

#define CPL_OPERATION CPL_IMAGE_FILTER_LINEAR_NOBPM
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute a linear filtering
  @param    in      The image to filter
  @param    ker     The kernel 
  @return   The filtered image or NULL in error case.
 
  This function computes the filtering without taking care of the bad pixels.
  It simply flags the input image bad pixels neighbours as bad to reflect the
  bad pixels filtering contamination.
  These can then be further reprocessed. 
  The bordures (depending of the size of the kernel) of the filtered image are 
  set to 0, and flagged as bad too.
  These can also be further reprocessed. 
  
  The input kernel should have an odd number of rows and columns.
  The maximum size of the kernel is 31x31.
  The output image has to be deallocated with cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
static cpl_image * cpl_image_filter_linear_nobpm(
        const cpl_image   *   in,
        const cpl_matrix  *   ker)
{
    cpl_image       *   image_out ;
    cpl_mask        *   bpm ;
    const double    *   ker_arr ;
    int                 nc, nr ;
    double              sum_pix ;
    double              ker_norm ;
    int                 hsx, hsy ;
    int                 curr_pos, im_pos, filt_pos;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in && ker, CPL_ERROR_NULL_INPUT, NULL) ;
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr & 1,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc <= 31, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr <= 31, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    ker_norm = cpl_matrix_norm_abs(ker);
    cpl_ensure(ker_norm > 0, CPL_ERROR_ILLEGAL_INPUT, NULL);

    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;
    
    /* Precompute inverse sum of filters coeffs */
    ker_norm = 1.0 / ker_norm ;
   
    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }

    /* Create the output bad pixel mask and flag bad pixels */
    if (in->bpm == NULL) {
        bpm = cpl_mask_new(in->nx, in->ny);
    } else {
        bpm = cpl_mask_duplicate(in->bpm);
        /* Expand any existing bad pixels */
        if (!cpl_mask_is_empty(bpm)) cpl_mask_dilation(bpm, ker);
    }
    
    /* Flag the image bordures as bad */
    cpl_mask_flag_borders(bpm, hsx, hsy) ;
    image_out->bpm = bpm;
    
    return image_out;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_MORPHO_NOBPM
/*----------------------------------------------------------------------------*/
/**
  @brief    Filter an image in spatial domain with a morpho kernel.
  @param    in          Image to filter.
  @param    ker         Filter definition.
  @return   1 newly allocated image or NULL in error case
   
  This function computes the filtering without taking care of the bad pixels.
  It simply flags the input image bad pixels neighbours as bad to reflect the
  bad pixels filtering contamination.
  These can then be further reprocessed. 
  The bordures (depending of the size of the kernel) of the filtered image are 
  set to 0, and flagged as bad too.
  These can also be further reprocessed. 

  The input filter is defined by a square matrix whose size is smaller than 
  5x5. The first element is applied to the minimum pixel in the neighborhood, 
  the second to the second-to-min, etc. and the last coefficient is applied 
  to the maximum pixel.
  The returned image is a newly allocated object, it must be deallocated using
  cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_DIVISION_BY_ZERO a division by 0 occurs
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
static cpl_image * cpl_image_filter_morpho_nobpm(
        const cpl_image   *   in,
        const cpl_matrix  *   ker)
{
    cpl_image       *   image_out ;
    cpl_mask        *   bpm ;
    double              sum_pix ;
    double              ker_norm ;
    int                 curr_pos, im_pos, filt_pos ;
    int                 hsx, hsy ;
    int                 nr, nc ;
    const double    *   ker_arr;
    double          *   pix_arr ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in && ker, CPL_ERROR_NULL_INPUT, NULL) ;
 
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1,   CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc <= 5,  CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nc == nr, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    ker_norm = cpl_matrix_norm_abs(ker);
    cpl_ensure(ker_norm > 0, CPL_ERROR_ILLEGAL_INPUT, NULL);
    
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;
    
    /* Precompute inverse sum of filters coeffs */
    ker_norm = 1.0 / ker_norm ;

    /* Allocate pix_arr */
    pix_arr = cpl_malloc(nr*nc*sizeof(double)) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_free(pix_arr) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    /* Free */
    cpl_free(pix_arr) ;

    /* Create the output bad pixel mask and flag bad pixels */
    if (in->bpm == NULL) {
        bpm = cpl_mask_new(in->nx, in->ny);
    } else {
        bpm = cpl_mask_duplicate(in->bpm);
        /* Expand any existing bad pixels */
        if (!cpl_mask_is_empty(bpm)) cpl_mask_dilation(bpm, ker);
    }

    /* Flag the image bordures as bad */
    cpl_mask_flag_borders(bpm, hsx, hsy) ;
    image_out->bpm = bpm ;

    return image_out;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_MEDIAN_NOBPM
/*----------------------------------------------------------------------------*/
/**
  @brief    Apply a spatial median filter to an image.
  @param    in  Image to filter
  @param    ker    The kernel
  @return   1 newly allocated image or NULL in error case
   
  This function computes the filtering without taking care of the bad pixels.
  It simply flags the input image bad pixels neighbours as bad to reflect the
  bad pixels filtering contamination.
  These can then be further reprocessed. 
  The bordures (depending of the size of the kernel) of the filtered image are 
  set to 0, and flagged as bad too.
  These can also be further reprocessed. 

  Apply a spatial median filter to an image, return a newly allocated image
  which must be deallocated using cpl_image_delete().
  This median only accepts 3x3 kernels with 1s everywhere for the moment.
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
static cpl_image * cpl_image_filter_median_nobpm(
        const cpl_image   *   in,
        const cpl_matrix  *   ker)
{
    cpl_image       *   out ;
    cpl_mask        *   bpm ;
    int                 curpos, im_pos, filt_pos, pix_arr_pos ;
    int                 nr, nc ;
    int                 hsx, hsy ;
    const double    *   ker_arr ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in,  CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(ker, CPL_ERROR_NULL_INPUT, NULL);
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nr*nc <= 81, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker) > 0, CPL_ERROR_ILLEGAL_INPUT, NULL);
        
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;
    
    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
  
    /* Create the output bad pixel mask and flag bad pixels */
    if (in->bpm == NULL) {
        bpm = cpl_mask_new(in->nx, in->ny);
    } else {
        bpm = cpl_mask_duplicate(in->bpm);
        /* Expand any existing bad pixels */
        if (!cpl_mask_is_empty(bpm)) cpl_mask_dilation(bpm, ker);
    }

    /* Flag the image bordures as bad */
    cpl_mask_flag_borders(bpm, hsx, hsy) ;
    out->bpm = bpm ;

    return out;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMAGE_FILTER_STDEV_NOBPM
/*----------------------------------------------------------------------------*/
/**
  @brief    Standard deviation filter
  @param    in      input image
  @param    ker        the kernel
  @return   a newly allocated image or NULL in error case

  This function computes the filtering without taking care of the bad pixels.
  It simply flags the input image bad pixels neighbours as bad to reflect the
  bad pixels filtering contamination.
  These can then be further reprocessed. 
  The bordures (depending of the size of the kernel) of the filtered image are 
  set to 0, and flagged as bad too.
  These can also be further reprocessed. 

  For each pixel, compute the standard deviation of a local zone.
  The returned image must be deallocated using cpl_image_delete().
  Images can be CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the kernel size is not as requested
  - CPL_ERROR_TYPE_MISMATCH if the passed image type is not supported
 */
/*----------------------------------------------------------------------------*/
static cpl_image * cpl_image_filter_stdev_nobpm(
        const cpl_image         *    in,
        const cpl_matrix        *    ker)
{
    cpl_image       *   out ;
    cpl_mask        *   bpm ;
    const double    *   ker_arr ;
    int                 hsx, hsy ;
    int                 nr, nc ;
    register double     pix ;
    double              sum ;
    double              sq_sum ;
    double              inv_surf ;
    int                 i, j, k, l ;

    /* Test entries */
    cpl_ensure(in && ker, CPL_ERROR_NULL_INPUT, NULL);
    
    /* Get kernel informations */
    nr = cpl_matrix_get_nrow(ker) ;
    nc = cpl_matrix_get_ncol(ker) ;
    ker_arr = cpl_matrix_get_data_const(ker) ;
    
    /* Test the kernel validity */
    cpl_ensure(nc & 1, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(nr & 1, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    cpl_ensure(cpl_matrix_norm_abs(ker) > 0, CPL_ERROR_ILLEGAL_INPUT, 
            NULL);
        
    /* Initialize */
    hsx = (nc-1) / 2 ;
    hsy = (nr-1) / 2 ;

    /* Set inv_srf */
    inv_surf = 1.00 / (double)(nr*nc) ;

    /* Switch on image type */
    switch (in->type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL);
    }
     
    /* Create the output bad pixel mask and flag bad pixels */
    if (in->bpm == NULL) {
        bpm = cpl_mask_new(in->nx, in->ny);
    } else {
        bpm = cpl_mask_duplicate(in->bpm);
        /* Expand any existing bad pixels */
        if (!cpl_mask_is_empty(bpm)) cpl_mask_dilation(bpm, ker);
    }

    /* Flag the image bordures as bad */
    cpl_mask_flag_borders(bpm, hsx, hsy) ;
    out->bpm = bpm ;
 
    return out;
}
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Set the borders of a mask to CPL_BINARY_1
  @param    in      input mask
  @param    size_x  the border size in x
  @param    size_y  the border size in y
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  
  The input mask is modified.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the specified size is too big in respect
    to the image
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code cpl_mask_flag_borders(
        cpl_mask    *       in,
        int                 size_x,
        int                 size_y)
{
    cpl_binary * data = cpl_mask_get_data(in);
    const int    nx   = cpl_mask_get_size_x(in);
    const int    ny   = cpl_mask_get_size_y(in);
    int          i, j;
    

    cpl_ensure_code(in, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(size_x < nx/2, CPL_ERROR_ILLEGAL_INPUT);
    cpl_ensure_code(size_y < ny/2, CPL_ERROR_ILLEGAL_INPUT);

    for (j=0 ; j<size_y ; j++) {
        for(i=0 ; i<nx ; i++) data[i+j*nx] = CPL_BINARY_1 ;
    }
    for(j=size_y ; j<ny-size_y ; j++) {
        for (i=0 ; i<size_x ; i++) data[i+j*nx] = CPL_BINARY_1 ;
        for (i=nx-size_x ; i<nx ; i++) data[i+j*nx] = CPL_BINARY_1 ;
    }
    for (j=ny-size_y ; j<ny ; j++) {
        for(i=0 ; i<nx ; i++) data[i+j*nx] = CPL_BINARY_1 ;
    }

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the sum of the absolute values of the matrix elements
  @param    self The non-zero matrix.
  @return   The norm or 0 on error
  @note     This function will set an error code and return 0
              if the norm is less than 1/FLT_MAX.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_DIVISION_BY_ZERO if the norm would be less than 1/FLT_MAX

 */
/*----------------------------------------------------------------------------*/
static double cpl_matrix_norm_abs(const cpl_matrix * self)
{

    const int      nr = cpl_matrix_get_nrow(self);
    const int      nc = cpl_matrix_get_ncol(self);
    const double * elements;
    double         norm = 0;
    int i;


    cpl_ensure(nr > 0, cpl_error_get_code(), 0);

    elements = cpl_matrix_get_data_const(self);

    for (i=0 ; i < nr * nc ; i++) norm += fabs(elements[i]);

    cpl_ensure(FLT_MAX * norm > 1, CPL_ERROR_DIVISION_BY_ZERO, 0);

    return norm;

}

#define CPL_CLASS CPL_CLASS_DOUBLE
#undef CPL_OPERATION
#include "cpl_image_filter_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_FLOAT
#undef CPL_OPERATION
#include "cpl_image_filter_body.h"
#undef CPL_CLASS
