/* $Id: cpl_imagelist_basic.c,v 1.37 2008/07/21 08:48:40 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/07/21 08:48:40 $
 * $Revision: 1.37 $
 * $Name: cpl-4_2_0b3 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>

#include <cxstrutils.h>

#include "cpl_memory.h"
#include "cpl_imagelist_basic.h"
#include "cpl_imagelist_io.h"
#include "cpl_image_basic.h"
#include "cpl_image_io.h"
#include "cpl_image_bpm.h"
#include "cpl_stats.h"
#include "cpl_tools.h"
#include "cpl_mask.h"
#include "cpl_imagelist_defs.h"
#include "cpl_image_defs.h"

/*-----------------------------------------------------------------------------
                                   Define
 -----------------------------------------------------------------------------*/

#define CPL_IMLIST_BASIC_OPER               0
#define CPL_IMLIST_BASIC_IMAGE_LOCAL        1
#define CPL_IMLIST_BASIC_TIME_MEDIAN        2
#define CPL_IMLIST_BASIC_TIME_MINMAX        3
#define CPL_IMLIST_BASIC_TIME_SIGCLIP       4
#define CPL_IMLIST_BASIC_SWAP_AXIS          5

/*----------------------------------------------------------------------------*/
/**
 * @defgroup cpl_imagelist_basic    Images lists basic routines
 *
 * This module provides basic functions to handle @em cpl_imagelist.
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_imagelist_basic.h"
 * @endcode
 */
/*----------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                        Private function prototypes
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                            Function codes
 -----------------------------------------------------------------------------*/

#define CPL_OPERATION CPL_IMLIST_BASIC_OPER
/*----------------------------------------------------------------------------*/
/**
  @brief    Add two image lists, the first one is replaced by the result.
  @param    in1        first input image list (modified)
  @param    in2        image list to add
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_add()

  The two input lists must have the same size, the image number n in the
  list in2 is added to the image number n in the list in1.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input images have different sizes
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_add(
        cpl_imagelist         *    in1,
        const cpl_imagelist   *    in2)
{
#define CPL_OPERATOR cpl_image_add
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Subtract two image lists, the first one is replaced by the result.
  @param    in1        first input image list (modified)
  @param    in2        image list to subtract
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_subtract() 
  @see      cpl_imagelist_add()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_subtract(
        cpl_imagelist         *    in1,
        const cpl_imagelist   *    in2)
{
#define CPL_OPERATOR cpl_image_subtract
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Multiply two image lists, the first one is replaced by the result.
  @param    in1        first input image list (modified)
  @param    in2        image list to multiply
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_multiply()
  @see      cpl_imagelist_add()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_multiply(
        cpl_imagelist         *    in1,
        const cpl_imagelist   *    in2)
{
#define CPL_OPERATOR cpl_image_multiply
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Divide two image lists, the first one is replaced by the result.
  @param    in1        first input image list (modified)
  @param    in2        image list to divide
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_divide()
  @see      cpl_imagelist_add()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_divide(
        cpl_imagelist         *    in1,
        const cpl_imagelist   *    in2)
{
#define CPL_OPERATOR cpl_image_divide
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMLIST_BASIC_IMAGE_LOCAL
/*----------------------------------------------------------------------------*/
/**
  @brief    Add an image to an image list.
  @param    imlist  input image list (modified)
  @param    img     image to add
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_add()

  The passed image is added to each image of the passed image list.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_add_image(
        cpl_imagelist   *    imlist,
        const cpl_image *    img)
{
#define CPL_OPERATOR cpl_image_add
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Subtract an image from an image list.
  @param    imlist    input image list (modified)
  @param    img        image to subtract
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_subtract()
  @see      cpl_imagelist_add_image()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_subtract_image(
        cpl_imagelist   *    imlist,
        const cpl_image *    img)
{
#define CPL_OPERATOR cpl_image_subtract
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Multiply an image list by an image.
  @param    imlist    input image list (modified)
  @param    img     image to multiply
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_multiply()
  @see      cpl_imagelist_add_image()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_multiply_image(
        cpl_imagelist   *    imlist,
        const cpl_image *    img)
{
#define CPL_OPERATOR cpl_image_multiply
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Divide an image list by an image.
  @param    imlist    input image list (modified)
  @param    img     image for division
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_divide()
  @see      cpl_imagelist_add_image()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_divide_image(
        cpl_imagelist   *    imlist,
        const cpl_image *    img)
{
#define CPL_OPERATOR cpl_image_divide
#include "cpl_imagelist_basic_body.h"
#undef CPL_OPERATOR
}
#undef CPL_OPERATION

/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise addition of a scalar to each image in the imlist
  @param    imlist  Imagelist to be modified in place.
  @param    addend  Number to add
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_image_add_scalar()
 
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_add_scalar(
        cpl_imagelist * imlist,
        double          addend)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code( !cpl_image_add_scalar(imlist->images[i], addend),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise subtraction of a scalar from each image in the imlist
  @param    imlist   Imagelist to be modified in place.
  @param    subtrahend  Number to subtract
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_imagelist_add_scalar()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_subtract_scalar(
        cpl_imagelist * imlist,
        double          subtrahend)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code(!cpl_image_subtract_scalar(imlist->images[i],subtrahend),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise multiplication of the imlist with a scalar
  @param    imlist  Imagelist to be modified in place.
  @param    factor     Number to multiply with
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_imagelist_add_scalar()
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_multiply_scalar(
        cpl_imagelist * imlist,
        double          factor)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code( !cpl_image_multiply_scalar(imlist->images[i], factor),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Elementwise division of each image in the imlist with a scalar
  @param    imlist  Imagelist to be modified in place.
  @param    divisor    Non-zero number to divide with
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_imagelist_add_scalar()
 
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_divide_scalar(
        cpl_imagelist * imlist,
        double          divisor)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code( !cpl_image_divide_scalar(imlist->images[i], divisor),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the elementwise logarithm of each image in the imlist
  @param    imlist  Imagelist to be modified in place.
  @param    base       Base of the logarithm.
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_image_logarithm()
 
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_logarithm(
        cpl_imagelist * imlist,
        double          base)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code( !cpl_image_logarithm(imlist->images[i], base),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the elementwise exponential of each image in the imlist
  @param    imlist  Imagelist to be modified in place.
  @param    base       Base of the exponential.
  @return   CPL_ERROR_NONE or the relevant the #_cpl_error_code_ on error
  @see      cpl_image_exponential()
 
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_exponential(
        cpl_imagelist * imlist,
        double          base)
{
    int          i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
        
    for (i=0 ; i < imlist->ni ; i++)
        cpl_ensure_code( !cpl_image_exponential(imlist->images[i], base),
                         cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Normalize each image in the list.
  @param    imlist  Imagelist to modify.
  @param    mode    Normalization mode.
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error
  @see      cpl_image_normalise()

  The list may be partly modified if an error occurs.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_normalise(
        cpl_imagelist *  imlist,
        cpl_norm         mode)
{
    int                 i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
    
    for (i=0 ; i<imlist->ni ; i++)
        cpl_ensure_code(!cpl_image_normalise(imlist->images[i], mode), 
                        cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Threshold all pixel values to an interval.
  @param    imlist               Image list to threshold.
  @param    lo_cut          Lower bound.
  @param    hi_cut          Higher bound.
  @param    assign_lo_cut   Value to assign to pixels below low bound.
  @param    assign_hi_cut   Value to assign to pixels above high bound.
  @return   CPL_ERROR_NONE or the relevant #_cpl_error_code_ on error
  @see      cpl_image_threshold()

  Threshold the images of the list using cpl_image_threshold()
  The input image list is modified.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_INCOMPATIBLE_INPUT if lo_cut is bigger than hi_cut
 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_imagelist_threshold(
        cpl_imagelist * imlist,
        double          lo_cut,
        double          hi_cut,
        double          assign_lo_cut,
        double          assign_hi_cut)
{
    int                     i ;

    /* Check inputs */
    cpl_ensure_code(imlist != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(lo_cut < hi_cut, CPL_ERROR_INCOMPATIBLE_INPUT) ;
    
    for (i=0 ; i<imlist->ni ; i++)
        cpl_ensure_code(!cpl_image_threshold(imlist->images[i], lo_cut, hi_cut,
                                             assign_lo_cut, assign_hi_cut),
                        cpl_error_get_code());

    return CPL_ERROR_NONE ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create a contribution map from the bad pixel maps of the images.
  @param    imlist  The imagelist
  @return   The contributions map (a CPL_TYPE_INT cpl_image) or NULL on error
  @see cpl_imagelist_is_uniform()

  The returned map counts for each pixel the number of good pixels in the list.
  The returned map has to be deallocated with cpl_image_delete().

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_image_new_from_accepted(const cpl_imagelist * imlist)
{
    int                 nx, ny ;
    cpl_mask        *   bpm ;
    cpl_mask        *   bpmap ;
    cpl_image       *   bpm_int ;
    cpl_image       *   cmap ;
    int                 i ;
    
    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);

    /* Get informations */
    bpmap = cpl_image_get_bpm(imlist->images[0]) ;
    nx = cpl_image_get_size_x(imlist->images[0]) ;
    ny = cpl_image_get_size_y(imlist->images[0]) ;

    /* Create the contribution map with the first image of the list */
    if (bpmap == NULL) 
        bpm = cpl_mask_new(nx, ny) ; 
    else 
        bpm = cpl_mask_duplicate(bpmap) ;
    cpl_ensure(bpm!=NULL, CPL_ERROR_ILLEGAL_INPUT, NULL);

    /* Convert the bad pixel map in an integer image */
    cmap = cpl_image_new_from_mask(bpm) ;
    cpl_mask_delete(bpm) ;
    cpl_ensure(cmap, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
        
    /* Loop on the images */
    for (i=1 ; i<imlist->ni ; i++) {
        bpmap = cpl_image_get_bpm(imlist->images[i]) ;

        /* Create the current bad pixels map */
        if (bpmap == NULL) 
            bpm = cpl_mask_new(nx, ny);
        else 
            bpm = cpl_mask_duplicate(bpmap) ;
        if (bpm == NULL) {
            cpl_image_delete(cmap) ;
            cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
        }
        /* Convert the bad pixel map in an integer image */
        bpm_int = cpl_image_new_from_mask(bpm) ;
        cpl_mask_delete(bpm) ;
        if (bpm_int == NULL) {
            cpl_image_delete(cmap) ;
            cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
        }
        /* Update the contrbution map with the current plane */
        if (cpl_image_add(cmap, bpm_int)!=CPL_ERROR_NONE) {
            cpl_image_delete(cmap) ;
            cpl_image_delete(bpm_int) ;
            cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
        }
        cpl_image_delete(bpm_int) ;
    }
    
    /* cmap counts the bad pixels, we want it to count the good pixels */
    cpl_image_multiply_scalar(cmap, -1.0) ;
    cpl_image_add_scalar(cmap, imlist->ni) ;

    return cmap ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Average an imagelist to a single image
  @param    imlist    the input images list
  @return   the average image or NULL on error case.
  @see cpl_imagelist_is_uniform()

  The returned image has to be deallocated with cpl_image_delete().

  The bad pixel maps of the images in the input list are taken into
  account, the result image pixels are flagged as rejected for those
  where there were no good pixel at the same position in the input image list.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_collapse_create(const cpl_imagelist * imlist)
{
    cpl_image          *    avg ;
    cpl_image          *    tmp_im ;
    cpl_image          *    contrib ;
    cpl_mask           *    contrib_bin ;
    int                     i ;
    
    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);
    
    /* Create avg with the first image */
    avg = cpl_image_duplicate(imlist->images[0]) ;
    cpl_image_fill_rejected(avg, 0.0) ;
    cpl_image_accept_all(avg) ;
        
    /* Add all images together */
    for (i=1 ; i<imlist->ni ; i++) {
        tmp_im = cpl_image_duplicate(imlist->images[i]) ;
        cpl_image_fill_rejected(tmp_im, 0.0) ;
        cpl_image_accept_all(tmp_im) ;
        cpl_image_add(avg, tmp_im) ;
        cpl_image_delete(tmp_im) ;
    }

    /* Compute the contribution map */
    if ((contrib = cpl_image_new_from_accepted(imlist)) == NULL) {
        cpl_image_delete(avg) ;
        cpl_ensure(0, cpl_error_get_code(), NULL) ;
    }

    /* Divide by the number of contributions */
    if (cpl_image_divide(avg, contrib) != CPL_ERROR_NONE) {
        cpl_image_delete(contrib) ;
        cpl_image_delete(avg) ;
        cpl_ensure(0, cpl_error_get_code(), NULL) ;
    }

    /* The avg bad pixel map is the map of non contributing pixels */
    contrib_bin = cpl_mask_threshold_image_create(contrib, -0.5, 0.5) ;
    cpl_image_delete(contrib) ;
    if (contrib_bin==NULL) {
        cpl_image_delete(avg) ;
        cpl_ensure(0, cpl_error_get_code(), NULL) ;
    }

    /* If there are no bad pixels */
    if (cpl_mask_is_empty(contrib_bin)) {
        cpl_image_accept_all(avg) ;
    } else {
        if (cpl_image_reject_from_mask(avg, contrib_bin)) {
            cpl_mask_delete(contrib_bin);
            cpl_image_delete(avg) ;
            cpl_ensure(0, cpl_error_get_code(), NULL) ;
        }
    }
    cpl_mask_delete(contrib_bin) ;

    /* Return */
    return avg ;    
}

#define CPL_OPERATION CPL_IMLIST_BASIC_TIME_MINMAX
/*----------------------------------------------------------------------------*/
/**
  @brief    Average with rejection an imagelist to a single image
  @param    imlist  the input images list
  @param    low     number of low rejected values
  @param    high    number of high rejected values
  @return   the average image or NULL on error case.

  The returned image has to be deallocated with cpl_image_delete().

  The bad pixel maps of the input frames are not taken into account, and 
  the one of the created image is empty.

  The input images are averaged, for each pixel position the low lowest pixels 
  and the high highest pixels are discarded for the average computation.

  The input image list can be of type CPL_TYPE_INT, CPL_TYPE_FLOAT and
  CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid or if
    the sum of the rejections is not lower than the number of images or if
    low or high is negative
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_collapse_minmax_create(
        const cpl_imagelist *   imlist,
        int                     low,
        int                     high)
{
    const cpl_image *   cur_ima ;
    cpl_image       *   out_ima ;
    cpl_vector      *   time_line ;
    double          *   ptime_line ;
    double              out_val ;
    int                 ni, nx, ny ;
    int                 i, j, k, l ; 
    
    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);
    cpl_ensure(low>=0 && high>=0, CPL_ERROR_ILLEGAL_INPUT, NULL);
    ni = cpl_imagelist_get_size(imlist) ;
    cpl_ensure(ni > low+high, CPL_ERROR_ILLEGAL_INPUT, NULL);
    
    cur_ima = cpl_imagelist_get_const(imlist, 0) ;
    nx = cpl_image_get_size_x(cur_ima) ;
    ny = cpl_image_get_size_y(cur_ima) ;

    /* Create the vector */
    time_line = cpl_vector_new(ni) ;
    ptime_line = cpl_vector_get_data(time_line) ;

    /* Switch on the data type */
    switch (cpl_image_get_type(cur_ima)) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_vector_delete(time_line) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    cpl_vector_delete(time_line) ;

    /* Return */
    return out_ima ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMLIST_BASIC_TIME_SIGCLIP
/*----------------------------------------------------------------------------*/
/**
  @brief    Average with sigma-clipping rejection an imagelist to a single image
  @param    imlist  the input images list
  @param    sigma   the sigma value
  @return   the average image or NULL on error case.

  The returned image has to be deallocated with cpl_image_delete().

  The bad pixel maps of the input frames are not taken into account, and 
  the one of the created image is empty.

  For each pixel position the pixels whose value is higher than
  mean+sigma*stdev or lower than mean-sigma*stdev are discarded for the 
  average computation. If all pixels are discarde, the reslting pixel is 0.
  Where mean is the average of the pixels at that position, and stdev is
  the standard deviation of the pixels at that position.

  The input image list can be of type CPL_TYPE_INT, CPL_TYPE_FLOAT and
  CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid or if
  sigma is negative
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_collapse_sigclip_create(
        const cpl_imagelist *   imlist,
        double                  sigma)
{
    const cpl_image *   cur_ima ;
    int                 ni, nx, ny ;
    cpl_vector      *   time_line ;
    double          *   ptime_line ;
    cpl_image       *   out_ima ;
    double              out_val ;
    int                 nb_val ;
    double              mean, stdev ;
    double              low_thresh, high_thresh ;
    int                 i, j, k ; 
    
    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);
    cpl_ensure(sigma>0, CPL_ERROR_ILLEGAL_INPUT, NULL);
    ni = cpl_imagelist_get_size(imlist) ;
    
    cur_ima = cpl_imagelist_get_const(imlist, 0) ;
    nx = cpl_image_get_size_x(cur_ima) ;
    ny = cpl_image_get_size_y(cur_ima) ;

    /* Create the vector */
    time_line = cpl_vector_new(ni) ;
    ptime_line = cpl_vector_get_data(time_line) ;

    /* Switch on the data type */
    switch (cpl_image_get_type(cur_ima)) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_vector_delete(time_line) ;
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }
    cpl_vector_delete(time_line) ;

    /* Return */
    return out_ima ;
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMLIST_BASIC_TIME_MEDIAN
/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the median of an image list to a single image
  @param    imlist    the input images list
  @return   the median image or NULL on error case.
  @see cpl_imagelist_is_uniform()

  The returned image has to be deallocated with cpl_image_delete().

  The bad pixel maps of the input frames are not taken into account, and 
  the one of the created image is empty.

  The input image list can be of type CPL_TYPE_INT, CPL_TYPE_FLOAT and
  CPL_TYPE_DOUBLE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the input image list is not valid
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_image * cpl_imagelist_collapse_median_create(const cpl_imagelist * imlist)
{
    cpl_image           *   median ;
    int                     nx, ny ;
    cpl_type                type ;
    int                     i, j, k ;
    
    /* Check inputs */
    cpl_ensure(imlist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(imlist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);
    nx = cpl_image_get_size_x(imlist->images[0]);
    ny = cpl_image_get_size_y(imlist->images[0]);
    type = cpl_image_get_type(imlist->images[0]);

    /* Create the output image */
    median = cpl_image_new(nx, ny, type) ;

    cpl_ensure(median, cpl_error_get_code(), NULL);

    /* Switch on the data type */
    switch (type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS

#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_image_delete(median);
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL) ;
    }

    return median ;    
}
#undef CPL_OPERATION

#define CPL_OPERATION CPL_IMLIST_BASIC_SWAP_AXIS
/*----------------------------------------------------------------------------*/
/**
  @brief    Swap the axis of an image list
  @param    ilist   The image list to swap
  @param    mode    The swapping mode      
  @return The swapped image list or NULL in error case
 
  This function is intended for users that want to use the cpl_imagelist
  object as a cube. Swapping the axis would give them access to the
  usual functions in the 3 dimensions. This has the cost that it
  duplicates the memory consumption, which can be a problem for big
  amounts of data. 

  Image list can be CPL_TYPE_INT, CPL_TYPE_FLOAT or CPL_TYPE_DOUBLE.
  The mode can be either CPL_SWAP_AXIS_XZ or CPL_SWAP_AXIS_YZ
  
  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if mode is not equal to one of the possible
    values or if the image list is not valid
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_swap_axis_create(
        const cpl_imagelist *   ilist,
        cpl_swap_axis           mode)    
{
    cpl_imagelist       *   swapped ;
    const cpl_image     *   image ;
    cpl_image           *   cur_ima ;
    const cpl_image     *   old_ima ;
    int                     ni, nx, ny ;
    cpl_type                type ;
    int                     recompute_bpm, nbpm ;
    cpl_mask            *   cur_mask ;
    cpl_binary          *   pcur_mask ;
    cpl_binary          *   pold_mask ;
    int                     i, j, k ;

    /* Check entries */
    cpl_ensure(ilist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(cpl_imagelist_is_uniform(ilist)==0, CPL_ERROR_ILLEGAL_INPUT,
               NULL);
    cpl_ensure(mode==CPL_SWAP_AXIS_XZ || mode==CPL_SWAP_AXIS_YZ, 
            CPL_ERROR_ILLEGAL_INPUT, NULL) ;

    /* Initialise */
    recompute_bpm = 0 ;
    image = cpl_imagelist_get_const(ilist, 0);
    ni = cpl_imagelist_get_size(ilist);
    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);
    type = cpl_image_get_type(image) ;

    /* Create the new image list */
    swapped = cpl_imagelist_new() ;

    /* Switch on image type */
    switch (type) {
#define CPL_CLASS CPL_CLASS_DOUBLE
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_FLOAT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
#define CPL_CLASS CPL_CLASS_INT
#include "cpl_imagelist_basic_body.h"
#undef CPL_CLASS
        default:
            cpl_ensure(0, CPL_ERROR_TYPE_MISMATCH, NULL);
    }

    /* HANDLE THE BAD PIXELS MASKS */
    /* Check if there are bad pixels in the input image list */
    for (i=0 ; i<ni ; i++) {
        image = cpl_imagelist_get_const(ilist, i) ;
        if (image->bpm != NULL) {
            recompute_bpm = 1 ;
            break ;
        }
    }

    /* Need to compute the new bad pixels masks */
    if (recompute_bpm == 1) {
        /* SWAP X <-> Z */
        if (mode == CPL_SWAP_AXIS_XZ) {
            for (i=0 ; i<nx ; i++) {
                cur_mask = cpl_mask_new(ni, ny) ;
                pcur_mask = cpl_mask_get_data(cur_mask) ;
                nbpm = 0 ;
                for (j=0 ; j<ni ; j++) {
                    image = cpl_imagelist_get_const(ilist, j) ;
                    if (image->bpm != NULL) {
                        pold_mask = cpl_mask_get_data(image->bpm) ;
                        for (k=0 ; k<ny ; k++) {
                            if (pold_mask[i+k*nx] == CPL_BINARY_1) {
                                pcur_mask[j+k*ni]=pold_mask[i+k*nx] ;
                                nbpm ++ ;
                            }
                        }
                    }
                }

                /* Put the new bpm in the new image */
                if (nbpm > 0) {
                    cpl_image * swap = cpl_imagelist_get(swapped, i) ;
                    cpl_image_reject_from_mask(swap, cur_mask) ;
                }
                cpl_mask_delete(cur_mask) ;
            }
        } else {
        /* SWAP Y <-> Z */
            for (i=0 ; i<ny ; i++) {
                cur_mask = cpl_mask_new(nx, ni) ;
                pcur_mask = cpl_mask_get_data(cur_mask) ;
                nbpm = 0 ;
                for (j=0 ; j<ni ; j++) {
                    image = cpl_imagelist_get_const(ilist, j) ;
                    if (image->bpm != NULL) {
                        pold_mask = cpl_mask_get_data(image->bpm) ;
                        for (k=0 ; k<nx ; k++) {
                            if (pold_mask[k+i*nx] == CPL_BINARY_1) {
                                pcur_mask[k+j*ni]=pold_mask[k+i*nx] ;
                                nbpm ++ ;
                            }
                        }
                    }
                }

                /* Put the new bpm in the new image */
                if (nbpm > 0) {
                    cpl_image * swap = cpl_imagelist_get(swapped, i) ;
                    cpl_image_reject_from_mask(swap, cur_mask) ;
                }
                cpl_mask_delete(cur_mask) ;
            }
        }
    }
    return swapped ;
}
#undef CPL_OPERATION

/**@}*/
