/* $Id: cpl_tools_body.h,v 1.5 2008/07/21 08:48:38 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2008/07/21 08:48:38 $
 * $Revision: 1.5 $
 * $Name: cpl-4_2_0b3 $
 */

/*----------------------------------------------------------------------------*/
/**
  @brief    Sort an array
  @param    pix_arr      array to sort
  @param    n            array size
  @param    reverse      flag indicating whether to sort ascending (zero) or
                         descending (non-zero)
  @param    stable       flag to indicate whether to guarantee stability at
                         the cost of execution time (if cpl_tools_sort_int is
                         O(n log n), this function would still be O(n log n))
  @param    sort_pattern resulting sort pattern
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE

  The heap sort algorithm used here
  - has bad cache performance, but
  - is worst case O(n log n)
  - is stable (meaning that the order of equal elements is conserved 
               indepent on the value of the reverse flag)
  - is in place

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT
  - CPL_ERROR_ILLEGAL_INPUT
 */
/*----------------------------------------------------------------------------*/
cpl_error_code
CONCAT2X(cpl_tools_sort_stable_pattern, CPL_TYPE_NAME)(
    CPL_TYPE const *pix_arr,
    int n,
    int reverse,
    int stable,
    int *sort_pattern)
{
    int i;  
    
    /* Check entries */
    cpl_ensure(pix_arr, CPL_ERROR_NULL_INPUT,
               CPL_ERROR_NULL_INPUT) ;
    
    cpl_ensure(sort_pattern, CPL_ERROR_NULL_INPUT,
               CPL_ERROR_NULL_INPUT) ;

    if (n == 0) return CPL_ERROR_NONE;

    for (i = 0; i < n; i++) {
        sort_pattern[i] = i;
    }

    /* 
     * Heap sort
     */
    for (i = n / 2 - 1; i >= 0; i--) {
        int done = 0;
        int root = i;
        int bottom = n - 1;
        
        while ((root*2 + 1 <= bottom) && (!done)) {
            int child = root*2 + 1;

            if (child+1 <= bottom) {
                if ((!reverse && CPL_TOOLS_SORT_LT(
                         pix_arr[sort_pattern[child]],
                         pix_arr[sort_pattern[child + 1]]))
                    ||
                    (reverse && CPL_TOOLS_SORT_LT(
                        pix_arr[sort_pattern[child + 1]],
                        pix_arr[sort_pattern[child]]))
                    ) {
                    child += 1;
                }
            }

            if ((!reverse && CPL_TOOLS_SORT_LT(
                    pix_arr[sort_pattern[root]],
                    pix_arr[sort_pattern[child]])) 
                    ||
                    (reverse && CPL_TOOLS_SORT_LT(
                    pix_arr[sort_pattern[child]],
                    pix_arr[sort_pattern[root]]))
                    ) {
                CPL_INT_SWAP(sort_pattern[root], sort_pattern[child])
                    root = child;
            }
            else {
                done = 1;
            }
        }
    }
    
    for (i = n - 1; i >= 1; i--) {
        int done = 0;
        int root = 0;
        int bottom = i - 1;
        CPL_INT_SWAP(sort_pattern[0], sort_pattern[i])
        
        while ((root*2 + 1 <= bottom) && (!done)) {
            int child = root*2 + 1;

            if (child+1 <= bottom) {
                if ((!reverse && CPL_TOOLS_SORT_LT(
                         pix_arr[sort_pattern[child]],
                         pix_arr[sort_pattern[child + 1]]))
                    ||
                    (reverse && CPL_TOOLS_SORT_LT(
                        pix_arr[sort_pattern[child + 1]],
                        pix_arr[sort_pattern[child]]))
                    ) {
                    child += 1;
                }
            }
            if ((!reverse && CPL_TOOLS_SORT_LT(
                    pix_arr[sort_pattern[root]],
                    pix_arr[sort_pattern[child]])) 
                    ||
                (reverse && CPL_TOOLS_SORT_LT(
                    pix_arr[sort_pattern[child]],
                    pix_arr[sort_pattern[root]]))
                ) {
                CPL_INT_SWAP(sort_pattern[root], sort_pattern[child])
                    root = child;
            }
            else {
                done = 1;
            }
        }
    }

    /* 
     * Enforce stability
     */
    if (stable) {
        for (i = 0; i < n; i++) {
            int j;
            j = i + 1;
            while(j < n &&
                  !CPL_TOOLS_SORT_LT(pix_arr[sort_pattern[i]],
                                     pix_arr[sort_pattern[j]]) &&
                  !CPL_TOOLS_SORT_LT(pix_arr[sort_pattern[j]],
                                     pix_arr[sort_pattern[i]])) {
                j++;
            }
            if (j - i > 1) {
                cpl_tools_sort_int(sort_pattern + i, j - i);
            }
            i = j - 1; 
        }
    }     

    return CPL_ERROR_NONE ;
}

#ifdef CPL_TYPE_IS_NUM
/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Compute the summed sample variance of an array
  @param    a     The array
  @param    n     The (non-negative) array size
  @param    pmean Iff non-NULL, *pmean is the mean (at no extra cost)
  @return   The summed sample variance, S(n) = sum((a_i-mean)^2) (i=1 -> n)

Math explanation for ticket DFS05126, written by Lander de Bilbao.


$$\sigma2 = \frac{\sum_{i=1}^N (x_i - \overline{x})2}{N-1}$$


We concentrate on how to compute

$$\sum_{i=1}^N (x_i - \overline{x})2$$

developed as follows

$$\sum_{i=1}^N (x_i - \overline{x})2 = \sum_{i=1}^N x_i2 - 2 \, \overline{x} \, \sum_{i=1}^N x_i + N \, \overline{x}2$$

as 

$$ \overline{x} = \frac{\sum_{i=1}^N x_i}{N}$$

then we have 

$$\sum_{i=1}^N (x_i - \overline{x})2 =  \sum_{i=1}^N x_i2 - N \, \overline{x}2$$


Now we look and see if it possible to, after doing this computation for N samples, add the contribution of a new sample

We call the new sample $x_{n+1}$, and the mean taken into account this new sample, $\overline{x}_{n+1}$. For clarity, we rewrite the previous equation with $\overline{x}$ renamed as $\overline{x}_n$

$$\sum_{i=1}^N (x_i - \overline{x}_n)2 =  \sum_{i=1}^N x_i2 - N \, \overline{x}_n2$$

We want to compute now

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2$$

Developed

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^{N+1} x_i2 - (N+1) \, \overline{x}_{n+1}2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - (N+1) \, \overline{x}_{n+1}2$$

as

$$\overline{x}_{n+1} = \frac{\sum_{i=1}^N x_i + x_{n+1}}{N+1} $$

$$\overline{x}_{n+1} = \frac{\overline{x}_n \, N + x_{n+1}}{N+1} $$

then we have

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - (N+1) \, ( \, \frac{\overline{x}_n \, N + x_{n+1}}{N+1} ) \, ^2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - \frac{( \, \overline{x}_n \, N + x_{n+1} ) \, ^2}{N+1}$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - \frac{( \, N2 \, \overline{x}_n2 + 2 \, N \, \overline{x}_n \, x_{n+1} + x_{n+1}2 \, )}{N+1}$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - \frac{N2}{N+1} \, \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} - \frac1{N+1} \, x_{n+1}2$$

we add in both parts of the equation

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 + \frac1{N+1} \, \overline{x}_n2 =  \sum_{i=1}^N x_i2 + x_{n+1}2 - \frac{N2}{N+1} \, \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} - \frac1{N+1} \, x_{n+1}2 + \frac1{N+1} \, \overline{x}_n2$$

we can now group some terms

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 + \frac1{N+1} \, \overline{x}_n2 =  \sum_{i=1}^N x_i2 - \frac{N2 - 1}{N+1} \, \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} + \frac{N}{N+1} \, x_{n+1}2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 + \frac1{N+1} \, \overline{x}_n2 =  \sum_{i=1}^N x_i2 - (N-1) \, \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} + \frac{N}{N+1} \, x_{n+1}2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 + \frac1{N+1} \, \overline{x}_n2 =  \sum_{i=1}^N x_i2 - N \, \overline{x}_n2 + \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} + \frac{N}{N+1} \, x_{n+1}2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 =  \sum_{i=1}^N x_i2 - N \, \overline{x}_n2 + \frac{N}{N+1} \, \overline{x}_n2 - \frac{N}{N+1} \, 2 \, \overline{x}_n \, x_{n+1} + \frac{N}{N+1} \, x_{n+1}2$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 = \sum_{i=1}^N (x_i - \overline{x}_n)2 + \frac{N}{N+1} \,( \, \overline{x}_n2 - 2 \, \overline{x}_n \, x_{n+1} + x_{n+1}2 )$$

$$\sum_{i=1}^{N+1} (x_i - \overline{x}_{n+1})2 = \sum_{i=1}^N (x_i - \overline{x}_n)2 + \frac{N}{N+1} \,( \, \overline{x}_n - x_{n+1} )2 $$

 */
/*----------------------------------------------------------------------------*/
double CONCAT2X(cpl_tools_get_variancesum, CPL_TYPE_NAME)(const CPL_TYPE * a,
                                                          int n, double * pmean)
{
    double varsum = 0.0;
    double mean = 0.0;
    int i;

    cpl_ensure(a != NULL, CPL_ERROR_NULL_INPUT, 0.0);
    cpl_ensure(n >= 0, CPL_ERROR_ILLEGAL_INPUT, 0.0);

    for (i=0; i < n; i++) {
        const double delta = (double)a[i] - mean;

        varsum += i * delta * delta / (i + 1.0);
        mean   += delta / (i + 1.0);
    }

    cpl_tools_add_flops( 1 + 6 * n ); /* Assume expression reuse */

    if (pmean != NULL) *pmean = mean;

    return varsum;
}

/*----------------------------------------------------------------------------*/
/**
  @internal
  @brief    Compute the sample variance of an array
  @param    a     The array
  @param    n     The (positive) array size
  @param    pmean Iff non-NULL, *pmean is the mean (at no extra cost)
  @return   The sample variance, S(n) = (1/n) sum((a_i-mean)^2) (i=1 -> n)
  @see cpl_tools_get_variancesum_double()
 */
/*----------------------------------------------------------------------------*/
double CONCAT2X(cpl_tools_get_variance, CPL_TYPE_NAME)(const CPL_TYPE * a,
                                                       int n, double * pmean)
{

    const double varsum
        = CONCAT2X(cpl_tools_get_variancesum, CPL_TYPE_NAME)(a, n, pmean);

    cpl_ensure(n > 0, CPL_ERROR_ILLEGAL_INPUT, 0.0);

    return varsum / n;
}

#endif
